# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

# builtins
import copy
import logging

# external
from mixbox.vendor.six import iteritems

# internal
from stixmarx import attrmap
from stixmarx import errors
from stixmarx import navigator
from stixmarx import utils
from stixmarx import xml

# Module-level logger
LOG = logging.getLogger(__name__)


class MarkingSerializer(object):
    """Enables the serialization of markable content by creating XPath
    expressions that assert the resulting XML. Provides support for global
    markings, that is, markings that cover the entire package. Component
    markings, these cover an entity (typically a TLO) and all of its
    descendants. Field markings provides the most granular approach to mark
    specific content like: individual elements, attributes or text of XML.

    Note:
        This class is only used internally with MarkingContainer.

    Attributes:
        container: A MarkingContainer object.
        nsmap: A dictionary with the namespace to prefix map of STIX, CybOX and
            MAEC.

    """
    def __init__(self, marking_container):
        self._container = marking_container
        self._nsmap = utils.load_nsmap()

    def _apply_global_markings(self):
        package = self._container.package
        global_markings = self._container._global_markings

        if not global_markings:
            return

        handling = utils.get_handling(package)

        for marking in global_markings:
            marking = copy.deepcopy(marking)
            marking.controlled_structure = xml.XPATH_GLOBAL_ALL_FIELDS

            handling.add_marking(marking)

    def _apply_markings_to_field(self, field, marking_info):
        """Resolves XPath and Handling for the marking. It covers component and
        field markings.

        Args:
            field: A `markable` entity.
            marking_info: Tuple, contains MarkingSpecification object and
                True/False value to indicate if marking applies to descendants.
        """
        for marking, descendants in marking_info:
            marking = copy.deepcopy(marking)
            marking.controlled_structure, handling =\
                self._find_path_and_handling(field, descendants)

            handling.add_marking(marking)

    def _apply_field_markings(self):
        field_markings = self._container._field_markings

        if not field_markings:
            return

        for field, markings_info in iteritems(field_markings):
            self._apply_markings_to_field(field, markings_info)

    def _apply_null_markings(self):
        package = self._container.package
        null_markings = self._container._null_markings

        if not null_markings:
            return

        handling = utils.get_handling(package)

        for marking in null_markings:
            marking = copy.deepcopy(marking)
            handling.add_marking(marking)

    def _apply_markings(self):
        self._apply_global_markings()
        self._apply_field_markings()
        self._apply_null_markings()

    def serialize_xml(self, *args, **kwargs):
        """
        Applies the MarkingSpecification objects from global_markings and
        field makings from the MarkingContainer.

        Args:
            *args: Arguments from ``stix.Entity.to_xml()``.
            **kwargs: Keyword arguments from ``stix.Entity.to_xml()``.

        Returns:
            stix.core.STIXPackage: A STIX Package with markings explicitly
            applied.
        """
        package = self._container.package
        self._apply_markings()
        return package.to_xml(*args, **kwargs)

    def serialize_dict(self, *args, **kwargs):
        """
        Applies the MarkingSpecification objects from global_markings and
        field makings from the MarkingContainer.

        Args:
            *args: Arguments from ``stix.Entity.to_dict()``.
            **kwargs: Keyword arguments from ``stix.Entity.to_dict()``.

        Returns:
            stix.core.STIXPackage: A STIX Package with markings explicitly
            applied.
        """
        package = self._container.package
        self._apply_markings()
        return package.to_dict(*args, **kwargs)

    def _find_path_and_handling(self, field, descendants):
        """Generates an XPath expression based on the field provided. It also
        resolves `Handling` to indicate where the marking will be stored.

        Args:
            field: A `markable` entity.
            descendants: A boolean value. If True the generated XPath covers
                descendants.

        Returns:
            tuple: A Tuple containing the control structure string for the
                MarkingSpecification and the Handling object where the Marking
                will be stored.

        Raises:
            SerializerFieldNotFoundError: When a field marking was not found
                after walking the object model.
        """
        handling = None
        xpath = None
        found = False

        for entity_path in navigator.iterpath(self._container.package):
            if field is entity_path[2]:
                found = True
                ancestors = entity_path[0]
                field_name = entity_path[1]
                field_value = entity_path[2]

                ancestors, xpath, handling = self._resolve_handling(
                        ancestors,
                        field_value
                )

                for entity in ancestors:
                    index = ancestors.index(entity)
                    prefix = self._nsmap.preferred_prefix_for_namespace(
                            entity._namespace
                    )

                    mapping = self._map_to_xml(
                            index,
                            ancestors,
                            descendants,
                            field_name
                    )

                    predicate = self._resolve_xpath_predicate(
                            index,
                            ancestors,
                            field_value
                    )

                    if (attrmap.is_attribute(mapping) or
                            attrmap.is_content(mapping) or
                            utils.is_node(mapping)):
                        xpath.append(mapping)
                    else:
                        step = xml.XPATH_STRUCTURE.format(
                                prefix=prefix,
                                nodename=mapping,
                                predicates=predicate
                        )

                        xpath.append(step)

                index = len(ancestors) - 1
                mapping = self._map_to_xml(index, ancestors, descendants)

                if not (attrmap.is_attribute(xpath[-1]) or
                        attrmap.is_content(xpath[-1])):
                    xpath.append(mapping)
                break

        if not found:
            error = "Could not generate an XPath for {0}".format(field)
            raise errors.SerializerFieldNotFoundError(entity=field,
                                                      message=error)

        xpath = xml.XPATH_SELECT_OPERATOR.join(xpath)

        if xml.XPATH_AXIS_DESCENDANT_OR_SELF_NODE in xpath and descendants:
            all_attrs = xpath + xml.XPATH_SELECT_OPERATOR + xml.XPATH_WILDCARD_ALL_ATTRS
            xpath = xml.XPATH_JOIN_OPERATOR.format(xpath, all_attrs)

        return xpath, handling

    def _map_to_xml(self, index, path, descendants, field_name=None):
        """Maps Python markable entities to their equivalent XML representation
        by matching the current.

        Args:
            index: An int, keeps the index of the current object to map.
            path: A List of mixbox.entities.Entity ancestors.
            descendants: A boolean value. If True, XPath selects descendants.
            field_name: A string. The last field to map.

        Returns:
            str: A string, result of mapping the markable entity to an XML
                representation.

        Raises:
            SerializerMappingError: When a field cannot be mapped from the
                Python object model to the corresponding XML representation.
        """
        if index + 1 == len(path):
            if field_name is None and descendants is False:
                return xml.XPATH_AXIS_SELF_NODE

            elif field_name is None and descendants is True:
                return xml.XPATH_AXIS_DESCENDANT_OR_SELF_NODE

            result = attrmap.xmlfield(path[index], field_name)

            if result is not None:
                return result

            error = "Cannot find mapping for field '{0}' in '{1}'.".format(
                field_name, type(path[index]))
            raise errors.SerializerMappingError(entity=path[index],
                                                message=error)

        elif path[index]._namespace in self._nsmap:
            result = None

            for properties in path[index].typed_fields_with_attrnames():
                attr = properties[0]

                val = getattr(path[index], attr)

                if (val is path[index + 1] or
                        (utils.is_sequence(val) and path[index + 1] in val)):
                    result = attrmap.xmlfield(path[index], attr)
                    break

            if result is not None:
                return result

        error = "Cannot find mapping for field {0} in {1}.".format(
                type(path[index + 1]), type(path[index]))
        raise errors.SerializerMappingError(entity=path[index], message=error)

    def _resolve_xpath_predicate(self, index, path, last_obj=None):
        """Determines how to resolve the xpath predicate by finding the index
        of the next node using the ancestor.

        Args:
            index: An int, keeps the index of ancestor.
            path: A List of mixbox.entities.Entity ancestors.
            last_obj: A mixbox.entities.Entity object. The last field to map.

        Returns:
            int: An integer indicating the index of the XML node being mapped.
        """
        if index + 1 == len(path):
            current = path[index]
            return self._find_index_of_seq(current, last_obj)

        else:
            ancestor = path[index]
            current = path[index + 1]
            return self._find_index_of_seq(ancestor, current)

    def _resolve_handling(self, path, field_value):
        """Resolves where to place the marking based on the following factors:

        If the ancestors of the field to be do not contain a ``Handling``
        element, place the marking in the STIX_Header. Otherwise place the
        marking in the top-most TLO or in itself when a marking is placed to a
        TLO.

        Based on the handling selection, it may be necessary to remove some
        unnecessary ancestors (for TLO cases) or change the XPath start point.

        Args:
            path: A List of mixbox.entities.Entity ancestors.
            field_value: A mixbox.entities.Entity object. The last field to map.

        Returns:
            tuple: Containing three elements: list of ancestors, the XPath
                starting position and a Handling object.
        """
        for entity in path:
            if utils.contains_handling(entity):
                xpath = [xml.XPATH_TLO_RELATIVE_START]
                handling = utils.get_handling(entity)
                path = path[path.index(entity):]
                break

            elif utils.is_stix_report(entity):
                xpath = [xml.XPATH_HEADER_RELATIVE_START]
                handling = utils.get_handling(entity)
                path = path[path.index(entity):]
                break

        else:
            if utils.contains_handling(field_value):
                xpath = [xml.XPATH_TLO_RELATIVE_START]
                handling = utils.get_handling(field_value)
                path = []
            elif utils.is_stix_report(field_value):
                xpath = [xml.XPATH_HEADER_RELATIVE_START]
                handling = utils.get_handling(field_value)
                path = []
            else:
                xpath = [xml.XPATH_HEADER_RELATIVE_START]
                handling = utils.get_handling(self._container.package)

        return path, xpath, handling

    def _find_index_of_seq(self, object_, to_find):
        """Finds the corresponding positional node location of an object that
        matches the XML model and XPath predicate.

        Since XPath indexing starts at 1, all values found have an offset of 1.

        Args:
            object_: Ancestor Entity. Parent of `to_find`.
            to_find: Current object to find positional location.

        Returns:
            int: By default 1 if the Entity is not a sequence or only one
                instance exist. Otherwise the index + 1 offset.
        """
        if utils.is_sequence(object_):
            if to_find in object_:
                return object_.index(to_find) + 1

        for properties in object_.typed_fields_with_attrnames():
            attr = properties[0]

            val = getattr(object_, attr)

            if utils.is_sequence(val) and to_find in val:
                try:
                    return val.index(to_find) + 1
                except KeyError:
                    for idx, value in enumerate(val, start=0):
                        if value is to_find:
                            return idx + 1

        return 1



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

from lxml import etree

# Namespaces
NS_XSI = "http://www.w3.org/2001/XMLSchema-instance"
NS_XML_SCHEMA = "http://www.w3.org/2001/XMLSchema"

# LXML QNAMES TAGS
TAG_XS_INCLUDE = "{%s}include" % NS_XML_SCHEMA
TAG_XS_IMPORT = "{%s}import" % NS_XML_SCHEMA
TAG_XSI_TYPE = "{%s}type" % NS_XSI
TAG_SCHEMALOCATION = "{%s}schemaLocation" % NS_XSI

# Global selectors
XPATH_GLOBAL_ALL_ELEMS = "//node()"
XPATH_GLOBAL_ALL_ATTRS = "//@*"
XPATH_GLOBAL_ALL_FIELDS = "{0} | {1}".format(
    XPATH_GLOBAL_ALL_ELEMS,
    XPATH_GLOBAL_ALL_ATTRS
)

# Selectors relative to a TLO (top-level object)
XPATH_TLO_RELATIVE_START = "../../.."

# Selectors relative to a marking at the STIX Header object.
XPATH_HEADER_RELATIVE_START = "../../../.."

# Axis for nodes
XPATH_AXIS_SELF_NODE = "self::node()"
XPATH_AXIS_DESCENDANT_OR_SELF_NODE = "descendant-or-self::node()"

# Operators
XPATH_JOIN_OPERATOR = "{0} | {1}"
XPATH_SELECT_OPERATOR = "/"

# Selectors
XPATH_WILDCARD_ALL_ATTRS = "@*"

# Node selection structure
XPATH_STRUCTURE = "{prefix}:{nodename}[{predicates}]"


def is_element(node):
    """Return True if the input `node` is an XML element node."""
    return etree.iselement(node)


def is_attribute(node):
    """Return True if the input `node` is an XML attribute node."""
    if hasattr(node, 'is_attribute'):
        return node.is_attribute
    return False


def is_content(node):
    """Return True if the input `node` is a text node."""
    return ((isinstance(node, etree._ElementStringResult) or
            isinstance(node, etree._ElementUnicodeResult)) and
            not is_attribute(node))


def is_typed(node):
    """Return True if the `node` has a an xsi:type attribute."""
    if hasattr(node, 'attrib'):
        return TAG_XSI_TYPE in node.attrib
    return False


def xsitype_ns(node):
    """Return the datatype namespace for the input xsi:typed node.

    Args:
        node: An lxml etree Element node.

    Returns:
        The value of the input `node` @xsi:type attribute or None if no
        @xsi:type attribute is found.
    """
    if not is_typed(node):
        return None

    xsi_type = node.attrib[TAG_XSI_TYPE]
    alias = xsi_type.split(":")[0]
    namespace = node.nsmap[alias]
    return namespace


def is_etree(tree):
    """Return ``True`` if `tree` is an lxml etree ElementTree object."""
    return isinstance(tree, etree._ElementTree)


def to_etree(doc, encoding=None):
    """Return the `doc` as an ElementTree object.

    Args:
        doc: A filename, file-like object, or etree object (Element/ElementTree).
        encoding: The file encoding.

    Returns:
        An etree ElementTree object.
    """
    if is_etree(doc):
        return doc
    elif etree.iselement(doc):
        return etree.ElementTree(doc)
    else:
        parser = get_xml_parser(encoding=encoding)
        return etree.parse(doc, parser=parser)


def root(doc, encoding=None):
    """Return the root node for the input XML `doc`.

    Args:
        doc: A read()-able XML document, etree ElementTree or Element object.
        encoding: The encoding of the input document.

    Returns:
        An lxml Element object for the root of the input `doc`.
    """
    tree = to_etree(doc, encoding)
    return tree.getroot()


def get_xml_parser(encoding=None):
    """Returns an ``etree.ETCompatXMLParser`` instance."""
    parser = etree.ETCompatXMLParser(
        huge_tree=True,
        remove_comments=True,
        strip_cdata=False,
        remove_blank_text=True,
        resolve_entities=False,
        encoding=encoding
    )

    return parser


def localname(node):
    """Return the XML localname for the input lxml `node`.

    Args:
        node: An lxml etree element/attribute node.

    Returns:
        A string XML localname.
    """
    if node is None:
        return None
    if is_attribute(node):
        return etree.QName(node.attrname).localname
    else:
        return etree.QName(node).localname


def find(node, localname):
    """Find the first element under `node` with a matching localname.

    Args:
        node: An lxml etree Element object.
        localname: An XML localname for an element.

    Returns:
        An lxml etree Element object if found or None if not found.
    """
    xmltag = "{*}%s" % localname
    return node.find(xmltag)


def findall(node, localname):
    """Find all elements under `node` with a matching localname.

    Args:
        node: An lxml etree Element object.
        localname: An XML localname for an element.

    Returns:
        An iterable collection of  lxml etree Element objects.
    """
    if localname == "text()":
        # Fixes @delimiter separated values. We need a _ElementStringResult in
        # order to perform the correct matching.
        return node.xpath(localname)

    xmltag = "{*}%s" % localname
    return node.findall(xmltag)


def findattr(node, attrname):
    """Find the attribute under `node` with a matching attribute name.

    Args:
        node: An lxml etree Element object.
        attrname: An XML attribute name.

    Returns:
        An lxml etree ElementStringResults object if found or None if not found.
    """
    if not attrname.startswith("@"):
        attrname = "@%s" % attrname
    return next(iter(node.xpath(attrname)), None)


def findtext(node):
    """Return the text() node for the input `node` or None if no text() is
    found.

    Args:
        node: An lxml etree Element node.
    """
    return next(iter(node.xpath("text()")), None)



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.


class MissingMarkingPathError(Exception):
    def __init__(self, message, entity):
        super(MissingMarkingPathError, self).__init__(message)
        self.entity = entity


class UnrecognizedMarkingPathError(Exception):
    def __init__(self, message, entity, path):
        super(UnrecognizedMarkingPathError, self).__init__(message)
        self.entity = entity
        self.path = path


class UnknownMarkingError(Exception):
    def __init__(self, message, marking):
        super(UnknownMarkingError, self).__init__(message)
        self.marking = marking


class DuplicateMarkingError(Exception):
    def __init__(self, message, entity, marking):
        super(DuplicateMarkingError, self).__init__(message)
        self.entity = entity
        self.marking = marking


class UnmarkableError(Exception):
    def __init__(self, message, entity):
        super(UnmarkableError, self).__init__(message)
        self.entity = entity


class InvalidRootError(Exception):
    def __init__(self, message, entity):
        super(InvalidRootError, self).__init__(message)
        self.entity = entity


class IdLookupError(Exception):
    def __init__(self, message, id_):
        super(IdLookupError, self).__init__(message)
        self.id_ = id_


class InvalidModeError(Exception):
    def __init__(self, message, found, expected):
        super(InvalidModeError, self).__init__(message)
        self.found = found
        self.expected = expected


class MarkingNotFoundError(Exception):
    def __init__(self, message, entity, marking):
        super(MarkingNotFoundError, self).__init__(message)
        self.entity = entity
        self.marking = marking


class MarkingPathNotEmpty(Exception):
    def __init__(self, message, marking):
        super(MarkingPathNotEmpty, self).__init__(message)
        self.marking = marking


class MarkingRemovalError(Exception):
    def __init__(self, message, entity, marking):
        super(MarkingRemovalError, self).__init__(message)
        self.entity = entity
        self.marking = marking


class SerializerMappingError(Exception):
    def __init__(self, message, entity):
        super(SerializerMappingError, self).__init__(message)
        self.entity = entity


class SerializerFieldNotFoundError(Exception):
    def __init__(self, message, entity):
        super(SerializerFieldNotFoundError, self).__init__(message)
        self.entity = entity



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

# stdlib
import datetime

# external
from mixbox import dates
from mixbox.vendor import six
from mixbox.vendor.six import text_type, binary_type

# internal
from stixmarx import utils

if six.PY3:
    long = int


class MarkableBool(int):
    """Python `bool` cannot be subclassed directly.
    This class subclasses `int` (just as `bool` does)
    and behaves like a `bool`.
    """

    def __init__(self, value):
        super(MarkableBool, self).__init__()
        self.__datamarkings__ = set()

    def __new__(cls, value):
        return int.__new__(cls, bool(value))

    def __repr__(self):
        return str(bool(self))


class MarkableInt(int):
    def __init__(self, value):
        super(MarkableInt, self).__init__()
        self.__datamarkings__ = set()

    def __new__(cls, value):
        return int.__new__(cls, value)


class MarkableLong(long):
    def __init__(self, value):
        super(MarkableLong, self).__init__()
        self.__datamarkings__ = set()

    def __new__(cls, value):
        return long.__new__(cls, value)


class MarkableFloat(float):
    def __init__(self, value):
        super(MarkableFloat, self).__init__()
        self.__datamarkings__ = set()

    def __new__(cls, value):
        return float.__new__(cls, value)


class MarkableDateTime(datetime.datetime):
    def __init__(self, value):
        super(MarkableDateTime, self).__init__()
        self.__datamarkings__ = set()

    def __new__(cls, value):
        dt = dates.parse_datetime(value)

        return datetime.datetime.__new__(
            cls,
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            microsecond=dt.microsecond,
            tzinfo=dt.tzinfo
        )


class MarkableDate(datetime.date):
    def __init__(self, value):
        super(MarkableDate, self).__init__()
        self.__datamarkings__ = set()

    def __new__(cls, value):
        dt = dates.parse_date(value)

        return datetime.date.__new__(
            cls,
            year=dt.year,
            month=dt.month,
            day=dt.day
        )


class MarkableBytes(binary_type):
    def __init__(self, value):
        super(MarkableBytes, self).__init__()
        self.__datamarkings__ = set()

    def __new__(cls, value):
        value = value or ""
        return binary_type.__new__(cls, value)


class MarkableText(text_type):
    def __init__(self, value):
        super(MarkableText, self).__init__()
        self.__datamarkings__ = set()

    def __new__(cls, value):
        value = value or ""
        return text_type.__new__(cls, value)


if six.PY2:
    TYPEMAP = {
        bool: MarkableBool,
        int: MarkableInt,
        float: MarkableFloat,
        long: MarkableLong,
        binary_type: MarkableBytes,
        text_type: MarkableText,
        datetime.datetime: MarkableDateTime,
        datetime.date: MarkableDate
    }
else:
    TYPEMAP = {
        bool: MarkableBool,
        int: MarkableInt,
        float: MarkableFloat,
        binary_type: MarkableBytes,
        text_type: MarkableText,
        datetime.datetime: MarkableDateTime,
        datetime.date: MarkableDate
    }


def is_castable(obj):
    """Return True if the input `obj` can be casted to a markable datatype.

    If `obj` is a list, this will check that each member of the list is
    castable.
    """
    if utils.is_list(obj):
        return all(type(x) in TYPEMAP for x in obj)
    else:
        return type(obj) in TYPEMAP


def cast(input_):
    """Convert the castable `input` object into a markable object.

    Args:
        input_: A castable object or a list of castable objects.

    Returns:
        A markable conversion of the `input` object. E.g., A `str` object will
        become a MarkableString object.
    """
    if utils.is_list(input_):
        return [cast(x) for x in input_]

    type_ = type(input_)
    klass = TYPEMAP[type_]
    return klass(input_)



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

"""
Defines a mapping of Python object properties to XML fieldnames.
"""

from stixmarx import fields
from stixmarx import utils


def is_attribute(fieldname):
    """Return True if the `fieldname` represents an XML attribute selector."""
    return fieldname.startswith("@")


def is_content(fieldname):
    """Return True if the `fieldname` represents an XML content selector."""
    return fieldname == "text()"


def is_element(fieldname):
    """Return True if the `fieldname` represents an XML element selector."""
    if is_attribute(fieldname):
        return False
    elif is_content(fieldname):
        return False
    else:
        return True


def xmlfield(entity, attr):
    """Return the XML field name for the mixbox Entity attribute.

    Args:
        entity: A mixbox Entity object.
        attr: An attribute name.

    Returns:
        An XML field selector for the associated Entity attribute. If no
        mapping exists, None is returned.
    """
    field_map = mapping(entity)
    return field_map.get(attr)


def mapping(entity):
    """Return a dictionary which maps the input `entity` attributes to
    XML field selectors.

    Args:
        entity: A mixbox Entity object.

    Returns:
        A dictionary which maps `entity` attributes to XML field selectors.
        If there is no mapping, an empty dictionary is returned.
    """
    klass = utils.fully_qualified_name(entity)
    nomap = {}

    if klass in fields.get_field_mappings():
        return fields.get_field_mappings().get(klass)
    else:
        return nomap



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

# stdlib
import logging

# mixbox
from mixbox import signals

# python-stix
from stix.core import STIXPackage
from stix.data_marking import MarkingSpecification

# stixmarx
from stixmarx import api
from stixmarx import xml
from stixmarx import attrmap
from stixmarx import markingmap

# stixmarx api
from stixmarx.api import types


# Module-level logger
LOG = logging.getLogger(__name__)


# Global constants
_ATTR_SOURCENODE = "__sourcenode__"
_ATTR_BINDING = "__binding__"
_NSMAP = {"marking": "http://data-marking.mitre.org/Marking-1"}


def _binding(entity):
    """Return the __binding__ attribute value on `entity` if found or None
    if not found.

    Args:
        entity: A mixbox Entity object.
    """
    if hasattr(entity, _ATTR_BINDING):
        return entity.__binding__
    return None


def _sourcenode(entity):
    """Return the source lxml node which `entity` was created from or None
    if not found.

    Args:
        entity: A mixbox Entity object.
    """
    binding = _binding(entity)

    if binding is None:
        return None
    elif hasattr(binding, _ATTR_SOURCENODE):
        return binding.__sourcenode__
    else:
        return None


class MarkingParser(object):
    """Parses STIX XML documents and decorates STIXPackage objects with
    field-level markings.

    Attributes:
        _encoding: The encoding of the input document.
        _root: The root lxml element of the parsed input document.
        _markingmap: A MarkingMap object which associates each element/attribute
            in the input document with the Marking elements which mark them.
        _entities: A list of mixbox Entity objects that were created during
            parse().
    """

    def __init__(self, root, encoding=None):
        self._encoding = encoding
        self._root = xml.root(root, encoding)
        self._markingmap = markingmap.build(self._root, encoding)
        self._entities = list()

        # Connect our mixbox signal receiver
        signals.connect("Entity.created.from_obj", self._handle_entity_created)

    def _handle_entity_created(self, entity, binding):
        """Receiver for the mixbox Entity.created.from_obj signal.

        Attach the `binding` object to `entity` via a __binding__ attribute
        and then store `entity` in the _entities list for later processing.

        This allows us to inspect the full ancestry of `entity` since
        __binding__ has a __sourcenode__ attribute containing the lxml node
        that the __binding__ was created from:

        entity.__binding__.__sourcenode__: API-->binding-->lxml node.

        Args:
            entity: A mixbox Entity object.
            binding: The generated binding object that the `entity` was
                created from.
        """
        entity.__binding__ = binding
        self._entities.append(entity)

    def _set_list_field_marking(self, entity, attr, specs):
        """Attach marking information to each item in the list found in the
        input ``entity.attr``.

        This will also cast each value in list to a "markable" type which
        subclasses the original type. This is done via the stixmarx.types module.

        Note:
            Unmarked items in the list will not be altered in any way.

        This method is required when the API has a list attribute that holds
        a list of builtin types. This occurs when the underlying schema defines
        a flat list of builtin XSD types(e.g., "Indicator/Alternative_ID", which
        is a 0..many of xs:string).

        Args:
            entity: A mixbox Entity object.
            attr: The name of an attribute on the `entity` that has a list
                value.
            specs: A mapping of lxml MarkingSpecificationType instance nodes
                to their corresponding python-stix MarkingSpecification objects.
        """
        valuelist = getattr(entity, attr)
        node = _sourcenode(entity)
        xmlfield = attrmap.xmlfield(entity, attr)
        xmlnodes = xml.findall(node, xmlfield)
        marked = [x for x in xmlnodes if x in self._markingmap]

        for idx, val in enumerate(valuelist):
            for xmlnode in marked:
                value = valuelist[idx]
                markings = (specs[x] for x in self._markingmap[xmlnode])

                markable = api.add_markings(value, markings)
                valuelist[idx] = markable  # Replace the value in the list

    def _set_field_marking(self, entity, attr, specs):
        """Convert the `attr` attribute on `entity` to a markable type,
        attach marking information to it, and replace the original value
        with the converted, marked value.

        Note:
            If the input entity.attr is not marked, this will not perform
            any conversion or replacement.

        Args:
            entity: A mixbox Entity object.
            attr: The name of an attribute on the `entity`.
            specs: A mapping of lxml MarkingSpecificationType instance nodes
                to their corresponding python-stix MarkingSpecification objects.
        """
        node = _sourcenode(entity)
        xmlfield = attrmap.xmlfield(entity, attr)

        if attrmap.is_attribute(xmlfield):
            xmlnode = xml.findattr(node, xmlfield)
        elif attrmap.is_content(xmlfield):
            xmlnode = xml.findtext(node)
        else:
            xmlnode = xml.find(node, xmlfield)

        # If the node was not marked, do not perform any conversion or replacement.
        if xmlnode not in self._markingmap:
            return

        value = getattr(entity, attr)
        markings = (specs[x] for x in self._markingmap[xmlnode])

        markable = api.add_markings(value, markings)
        setattr(entity, attr, markable)

    def _process_attrs(self, entity, specs):
        """Attach marking information to all all non-Entity fields on `entity`
        which have been marked.

        Note:
            This will convert non-Entity attribute values to markable datatypes.
            E.g, "str" will become "MarkableBytes".

        Args:
            entity: A mixbox Entity object.
            specs: specs: A mapping of lxml MarkingSpecificationType instance
                nodes to their corresponding python-stix MarkingSpecification
                objects.
        """
        node = _sourcenode(entity)

        if node is None:
            return

        for attr in attrmap.mapping(entity):
            value = getattr(entity, attr)

            if not types.is_castable(value):
                continue
            elif isinstance(value, list):
                self._set_list_field_marking(entity, attr, specs)
            else:
                self._set_field_marking(entity, attr, specs)

    def _process_entity(self, entity, specs):
        """Attach marking information to the input `entity` if the entity has
        been marked.

        Args:
            entity: A mixbox Entity object.
            specs: specs: A mapping of lxml MarkingSpecificationType instance
                nodes to their corresponding python-stix MarkingSpecification
                objects.
        """
        node = _sourcenode(entity)

        if node not in self._markingmap:
            return

        markings = (specs[x] for x in self._markingmap[node])
        api.add_markings(entity, markings)

    def _get_marking_specification_nodemap(self):
        """Build a dictionary which maps MarkingSpecificationType XML instance
        nodes (lxml Elements) to their associated python-stix
        MarkingSpecification objects.

        Returns:
            A dictionary. The keys are lxml element nodes, and the values are
            python-stix MarkingSpecification objects.
        """
        specmap = {}
        parsed = self._entities
        specs = (x for x in parsed if isinstance(x, MarkingSpecification))

        for entity in specs:
            source = _sourcenode(entity)  # TODO (bworrell): check for None?
            specmap[source] = entity

        return specmap

    def _cleanup(self, entity):
        """Remove the __binding__ attribute we attached to the entity during
        parse. This will help reduce the memory footprint.

        Args:
            entity: A mixbox Entity object.
        """
        if _binding(entity):
            del entity.__binding__

    def _process_markings(self):
        """Attach marking information to each mixbox Entity that was created
        during ``parse()``.

        Once processed, the source __binding__ and __sourcenode__ will be
        removed from the parsed entity.

        Note:
            If a parsed Entity is not marked, no marking information or
            attributes will be attached to it.
        """
        specmap = self._get_marking_specification_nodemap()

        for entity in self._entities:
            self._process_entity(entity, specmap)
            self._process_attrs(entity, specmap)
            self._cleanup(entity)

    def parse(self):
        """Parse a STIX document and evaluate data markings found in the
        document. All marked fields will have marking information attached
        and can be queried via MarkingContainer interfaces.

        Returns:
            A STIXPackage object.
        """
        self._entities = list()  # Reset this in case of multiple parse() calls.

        # Parse the STIX Package.
        package = STIXPackage.from_xml(
            xml_file=self._root,
            encoding=self._encoding
        )

        # Attach marking information to all marked fields.
        self._process_markings()

        # Return our parsed STIXPackage.
        return package


def parse_xml(xml_input, encoding=None):
    parser = MarkingParser(root=xml_input, encoding=encoding)
    return parser.parse()



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

"""
This version of the API can operate on major components and fields of
STIX, CybOX and MAEC entities. Acting on anything not classified as an
"entity" per the above definition is not supported and will raise an exception.
When API functions say they accept a "markings" object, it means they accept a
MarkingSpecification object, not a MarkingStructure. If you have a
MarkingStructure you wish to use to mark something, you must first place it
inside a MarkingSpecification object and supply that object to the API.
"""

# stdlib
import collections
import itertools

# internal
from stixmarx import api
from stixmarx import errors
from stixmarx import navigator
from stixmarx import serializer
from stixmarx import utils

__all__ = ['MarkingContainer']


class MarkingContainer(object):
    """Enables the operation of data markings on STIX, CybOX and MAEC objects.
    
    A MarkingContainer provides interfaces for applying, accessing, clearing
    and removing data marking information from its wrapped STIX Package. A
    MarkingContainer has methods for processing marking information and
    serialization which translate the marked object model into XPath
    controlled structures.

    Note:
        A MarkingContainer should not be created directly. Instead, use
        stixmarx.parse() or stixmarx.new() to create MarkingContainer instances.

    Attributes:
        package (stix.core.STIXPackage): The package object (from python-stix)
            wrapped by this container.
        global_markings (list of MarkingSpecification): List of markings that
            apply to the container `package` as a whole (every descendant).
        field_markings (dict, maps markable objects to MarkingSpecification):
            Dictionary where keys are `markable` entities and the values are a
            `list` that contain tuples of MarkingSpecification and descendants
            (True/False) option.
        null_markings: (list of MarkingSpecification): List of markings that
            apply to this container but, will NOT mark anything inside. This
            means, no controlled structure will be resolved for this objects.
    """

    def __init__(self, package):
        """Initialize a MarkingContainer object.

        Args:
            package: A stix.core.STIXPackage object.
        """
        self._field_markings = collections.defaultdict(list)
        self._global_markings = []
        self._null_markings = []

        self._package = package

    def _reset_collections(self):
        self._field_markings = collections.defaultdict(list)
        self._global_markings = []
        self._null_markings = []

    @property
    def global_markings(self):
        """Return the global markings that have been set via add_global().

        Note:
            This property DOES NOT return markings that were applied by
            MarkingParser (even markings that were applied to all nodes in the
            parsed document).

        Returns:
            tuple: Tuple containing MarkingSpecification objects.
        """
        return tuple(self._global_markings)

    @property
    def null_markings(self):
        """Return the null markings that have been set via add_markings().
        Where `markable` is None.

        Note:
            This property DOES NOT return markings that were applied by
            MarkingParser.

        Returns:
            tuple: Tuple containing MarkingSpecification objects.
        """
        return tuple(self._null_markings)

    @property
    def field_markings(self):
        """Return the field markings that have been set via add_marking().

        Note:
            This property DOES NOT return markings that were applied by
            MarkingParser.

        Returns:
            dict: Dictionary where keys are `markable` entities and values are
            a list of tuples with MarkingSpecification objects and their
            corresponding (True/False) descendants option.
        """
        return dict(self._field_markings)

    @property
    def package(self):
        """Package wrapped by this MarkingContainer"""
        return self._package

    def _add_descendants(self, markable, marking):
        """Apply marking to `markable` in-place to descendants."""
        for descendant in navigator.iterwalk(markable):
            descendant = api.add_marking(descendant, marking)

    def _remove_descendants(self, markable, marking):
        """Remove marking from the `markable` descendants."""
        for descendant in navigator.iterwalk(markable):
            if api.contains_marking(descendant, marking):
                api.remove_marking(descendant, marking)

    def _clear_descendants(self, markable):
        """Clear markings from the `markable` descendants."""
        for descendant in navigator.iterwalk(markable):
            api.clear_markings(descendant)

    def _get_descendants(self, markable):
        """Return unique markings from the `markable` descendants."""
        uniques = set()
        for descendant in navigator.iterwalk(markable):
            markings = api.get_markings(descendant)
            if markings:
                uniques.update(markings)

        return uniques

    def _remove_marking_specification(self, marking):
        """Removes MarkingSpecification object from the object model."""
        for descendant in navigator.iterwalk(self._package):
            if hasattr(descendant, "handling") and descendant.handling:
                if marking in descendant.handling:
                    descendant.handling.remove(marking)
                    return

        msg = "Unable to remove marking from document: marking not found."
        raise errors.MarkingNotFoundError(message=msg, entity=self.package,
                                          marking=marking)

    def _assert_unique_marking(self, markable, marking):
        """Assert that the input `markable` object is not marked by `marking`.
        If `markable` is already marked by `marking`, raise an error.

        Raises:
            errors.DuplicateMarkingError: If `markable` is already marked by
                `marking`.
        """
        is_duplicate = self.is_marked(markable, marking=marking)

        if not is_duplicate:
            return

        msg = ("The entity is already marked with this marking or an "
               "equivalent marking.")

        raise errors.DuplicateMarkingError(message=msg, entity=markable,
                                           marking=marking)

    def add_marking(self, markable, marking, descendants=False):
        """Add the `marking` to the `markable` field/object. If `markable`
        is a built-in immutable Python type, it will be coerced into a
        stixmarx.api.types datatype.

        Note:
            The add_marking() function may not always be able to apply the
            markings in-place. Users should set the input field to the return
            object after calling add_marking().

        Note:
            Use this method to apply null markings. This is, markings that are
            present within the document but, do not apply to any field. The
            `markable` parameter MUST be None.

        Example:
            >>> print type(indicator.title)
            <type 'str'>
            >>> marked_title = add_marking(indicator.title, marking)
            >>> print type(marked_title)
            <class 'stixmarx.api.types.MarkableBytes'>
            >>> indicator.title = marked_title  # set the title to the return value

        Example:
            >>> print type(indicator.timestamp)
            <type 'datetime.datetime'>
            >>> marked_timestamp = add_marking(indicator.timestamp, marking)
            >>> print type(marked_timestamp)
            <class 'stixmarx.api.types.MarkableDateTime'>
            >>> indicator.timestamp = marked_timestamp # set timestamp to the return value

        Example:
            >>> print type(indicator)
            <class 'stix.indicator.indicator.Indicator'>
            >>> marked_indicator = add_marking(indicator, marking, descendants=True) # The equivalent of a component marking
            >>> print type(marked_indicator)
            <class 'stix.indicator.indicator.Indicator'>
            >>> indicator = marked_indicator


        Args:
            markable: An object to mark (e.g., an Indicator.title string).
            marking: A python-stix MarkingSpecification object.
            descendants: If true, add the marking to all descendants
                `markable`.

        Returns:
            The `markable` object with data marking information attached. If
            `markable` is a built-in immutable Python type (e.g., str), it will
            be changed to a stixmarx.api.types datatype.

        Raises:
            UnmarkableError: If `markable` is a STIXPackage object.
            DuplicateMarkingError: If `markable` is already marked by
                `marking`.
            MarkingPathNotEmpty: If `marking` controlled_structure is set.
        """
        utils.check_marking(marking)
        utils.check_empty_marking(marking)

        # Handles null marking case.
        if markable is None:
            if marking not in self._null_markings:
                self._null_markings.append(marking)
                return
            else:
                msg = ("The marking is already present in the null_markings"
                       " internal collection.")

                raise errors.DuplicateMarkingError(message=msg,
                                                   entity=self.package,
                                                   marking=marking)

        self._assert_unique_marking(markable, marking)

        # API call before to avoid duplicates.
        marked = api.add_marking(markable, marking)

        if not utils.is_package(markable):
            if all((marking, descendants) != mark
                   for mark in self._field_markings[marked]):

                if descendants:
                    self._add_descendants(markable, marking)

                # Store marking and descendants option tied.
                self._field_markings[marked].append((marking, descendants))
                return marked

            msg = ("The marking is already present in the field_markings"
                   " internal collection.")

            raise errors.DuplicateMarkingError(message=msg,
                                               entity=self.package,
                                               marking=marking)

        msg = "Cannot mark STIX Package: use add_global()"
        raise errors.UnmarkableError(entity=markable, message=msg)

    def add_global(self, marking):
        """Add the `marking` MarkingSpecification object to the set of
        globally applicable markings (markings that apply to this container's
        package and all of its descendants).

        Markings added here will be included in the set returned from
        get_markings() for any valid field.

        Args:
            marking: A MarkingSpecification object.

        Raises:
            TypeError: If `marking` is not a MarkingSpecification object.
            MarkingPathNotEmpty: If `marking` controlled_structure is set.
            DuplicateMarkingError: If `marking` is already present in
                `global_markings` collection.
        """
        utils.check_marking(marking)
        utils.check_empty_marking(marking)

        if marking not in self._global_markings:
            self._global_markings.append(marking)
            return

        msg = ("The marking is already present in the global_markings"
               " internal collection.")

        raise errors.DuplicateMarkingError(message=msg,
                                           entity=self.package,
                                           marking=marking)

    def flush(self):
        """Flush markings onto package object.

        Markings are buffered in the MarkingContainer until explicitly flushed
        out to the MarkingContainer's package through this method.

        Note:
            The global and fields collection will reset after this call.

        Returns:
            stix.core.STIXPackage: A STIX Package with all makings explicitly
                applied from the container.
        """
        writer = serializer.MarkingSerializer(marking_container=self)
        writer._apply_markings()

        # Reset the collections so we don't return duplicates
        self._reset_collections()

        return self.package

    def get_markings(self, markable, descendants=False, null_markings=False):
        """Return the markings associated with the input `markable` object.

        Note:
            This will include any global markings that have not been explicitly
            applied to this field.

        Args:
            markable: A markable object (e.g., indicator.title).
            descendants: If True, return markings which apply to the input
                field and all of its descendants.
            null_markings: If True, return internal markings that do NOT apply
                to any markable. This null markings have not been explicitly
                set to the wrapped document. Use utils.get_null_markings(...)
                to find null markings that have been explicitly set in the
                document.

        Returns:
            list: A list of MarkingSpecification objects.
        """
        item_markings = api.get_markings(markable)
        descendant_markings_collection = ()
        null_markings_collection = ()

        if descendants:
            descendant_markings_collection = self._get_descendants(markable)

        if null_markings:
            null_markings_collection = self._null_markings

        all_markings = itertools.chain(
            self._global_markings,
            item_markings,
            descendant_markings_collection,
            null_markings_collection
        )

        return list(set(all_markings))

    def is_marked(self, markable, marking=None, descendants=False):
        """Return True if `markable` contains marking information.

        Args:
            markable: An markable object.
            marking: A MarkingSpecification object.
            descendants: If set, inspect descendant fields for marking
                information.

        Raises:
            UnmarkableError: If `markable` is not an markable entity.
            UnknownMarkingError: If `marking` is not a MarkingSpecification
                object.

        Returns:
            bool: True under the following conditions: if `markable` contains
                marking information, if `markable` is marked by `marking`, if
                `markable` descendants contain markings or if global markings
                have been added through add_global(). Otherwise False.
        """
        if api.is_markable(markable):
            markings = self.get_markings(markable, descendants)

            if marking is not None:
                utils.check_marking(marking)
                return marking in markings

            return bool(markings)

        msg = "Could not verify markings to unmarkable entity."
        raise errors.UnmarkableError(entity=markable, message=msg)

    def remove_marking(self, markable, marking, descendants=False):
        """Remove the `marking` MarkingSpecification from `markable`.

        Note:
            Use remove_global to remove globally applied markings.

        Args:
            markable: An object which contains data markings.
            marking: A MarkingSpecification object.
            descendants: If True, remove `marking` from any descendants.

        Raises:
            UnmarkableError: If `markable` is not an markable entity.
            MarkingNotFoundError: If `markable` (or descendant of `markable` if
                `descendants` is True) is marked by `marking`. If marking was
                not found in the internal marking collection.
            MarkingRemovalError: If `marking` is inherited from an ancestor OR
                if `markable` is STIXPackage object.
            UnknownMarkingError: If `marking` is not a MarkingSpecification
                object.
        """
        utils.check_marking(marking)

        # Handles null marking case.
        if markable is None:
            # Attempt to remove null marking from internal collection.
            if marking in self._null_markings:
                try:
                    self._null_markings.remove(marking)
                    return
                except (AttributeError, ValueError):
                    msg = ("Unable to remove marking from internal null"
                           " marking collection. Marking not found.")
                    raise errors.MarkingNotFoundError(entity=markable,
                                                      message=msg,
                                                      marking=marking)
            # Attempt to remove null marking from wrapped STIX Package.
            else:
                self._remove_marking_specification(marking)
                return

        if api.is_markable(markable):

            # Attempt to remove field-level marking from internal collection.
            if markable in self._field_markings:
                field_markings = self._field_markings[markable]

                try:
                    field_markings.remove((marking, descendants))
                    self._field_markings[markable] = field_markings

                    if api.contains_marking(markable, marking):
                        api.remove_marking(markable, marking)

                        if descendants:
                            self._remove_descendants(markable, marking)

                    return
                except (AttributeError, ValueError):
                    msg = ("Unable to remove marking from internal field-level"
                           " marking collection. Marking not found.")
                    raise errors.MarkingNotFoundError(entity=markable,
                                                      message=msg,
                                                      marking=marking)

            # Attempt to remove marking from wrapped STIX Package.
            elif api.contains_marking(markable, marking):

                if utils.is_package(markable):
                    msg = ("Cannot remove marking from STIX Package: use"
                           " remove_global()")
                    raise errors.MarkingRemovalError(entity=markable,
                                                     message=msg,
                                                     marking=marking)

                for mark in navigator.iterwalk(self._package):
                    if api.contains_marking(mark, marking):
                        if mark is markable:
                            api.remove_marking(markable, marking)

                            if descendants:
                                self._remove_descendants(markable, marking)

                            self._remove_marking_specification(marking)

                            return
                        else:
                            msg = ("Unable to remove marking. Marking is "
                                   "inherited from an ancestor.")
                            raise errors.MarkingRemovalError(message=msg,
                                                             entity=markable,
                                                             marking=marking)

            msg = "Unable to remove marking from markable. Marking not found."
            raise errors.MarkingNotFoundError(entity=markable, message=msg,
                                              marking=marking)

        msg = "Could not remove markings to unmarkable entity."
        raise errors.UnmarkableError(entity=markable, message=msg)

    def clear_markings(self, markable, descendants=False):
        """Remove all markings from the `markable` marked object.

        Args:
            markable: A marked object (e.g., indicator.title)
            descendants: If True, clear markings from `markable` and
                its descendants.

        Raises:
            UnmarkableError: If `markable` is not an markable entity.
        """
        if api.is_markable(markable):
            api.clear_markings(markable)

            if descendants:
                self._clear_descendants(markable)

            return

        msg = "Could not clear markings to unmarkable entity."
        raise errors.UnmarkableError(entity=markable, message=msg)

    def remove_global(self, marking):
        """Remove a globally-applied marking from the internal collection or
        from a parsed document that contain globally-applied markings.

        Args:
          marking (MarkingSpecification): marking to un-apply from global

        Raises:
            MarkingNotFoundError: If `marking` is not found in the
                global markings registry.
        """
        utils.check_marking(marking)

        # Attempt to remove marking from internal collection
        if marking in self._global_markings:
            try:
                self._global_markings.remove(marking)
                return
            except (AttributeError, ValueError):
                pass

        # Attempt to remove marking from wrapped STIX package.
        else:
            global_markings = self.get_markings(self._package)

            if marking in global_markings:
                try:
                    api.remove_marking(self._package, marking)
                    self._remove_descendants(self._package, marking)
                    self._remove_marking_specification(marking)
                    return
                except (AttributeError, ValueError):
                    pass

        msg = ("MarkingSpecification not found in global markings internal"
               " collection or in the wrapped STIX Package.")
        raise errors.MarkingNotFoundError(marking=marking, message=msg,
                                          entity=self._global_markings)

    def to_xml(self, *args, **kwargs):
        """Return an XML string of the STIX package represented by the Package
        object, with markings applied through the MarkingContainer.
        
        Uses the same arguments as ``stix.Entity.to_xml()``.
        """
        writer = serializer.MarkingSerializer(marking_container=self)
        xml_out = writer.serialize_xml(*args, **kwargs)

        # Reset the collections so duplicates aren't returned
        self._reset_collections()

        return xml_out

    def to_dict(self, *args, **kwargs):
        """Return a dictionary of the STIX Package represented by the Package
        object, with markings applied through the MarkingContainer.

        Uses the same arguments as ``stix.Entity.to_dict()``.
        """
        writer = serializer.MarkingSerializer(marking_container=self)
        dict_out = writer.serialize_dict(*args, **kwargs)

        # Reset the collections so duplicates aren't returned
        self._reset_collections()

        return dict_out



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

# builtin
import logging

# internal
from stixmarx.version import __version__  # noqa

formatter = logging.Formatter("[%(name)s] [%(levelname)-7s] [%(asctime)s] %(message)s")

# Console Handler for stixmarx messages
ch = logging.StreamHandler()
ch.setFormatter(formatter)

# Module-level logger
log = logging.getLogger(__name__)
log.addHandler(ch)
log.propagate = False


def parse(xml_input, encoding=None):
    from stixmarx import parser
    from stixmarx import container

    stix_package = parser.parse_xml(xml_input, encoding)
    marking_container = container.MarkingContainer(stix_package)

    return marking_container


def new():
    from stixmarx import container
    from stix.core import STIXPackage

    return container.MarkingContainer(STIXPackage())



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

# builtins
import contextlib
import copy
import datetime
import importlib
import inspect
import logging
import os
import pkgutil

# external
from mixbox import entities
from mixbox.vendor.six import itervalues, iteritems, iterkeys, string_types

# internal
from stixmarx import errors
from stixmarx import version
from stixmarx import xml

# Module-level logger
LOG = logging.getLogger(__name__)

# Lazy python-stix imports
stix = None
stix_core = None
stix_campaign = None
stix_common = None
stix_coa = None
stix_dm = None
stix_et = None
stix_incident = None
stix_indicator = None
stix_report = None
stix_ta = None
stix_ttp = None
stix_utils = None
stix_parser = None
stix_nsparser = None

# Lazy python-cybox imports
cybox = None
cybox_core = None
cybox_common = None
cybox_nsparser = None

# Lazy mixbox imports
mixbox = None
mixbox_entities = None

# Lazy python-maec imports
maec = None
maec_nsparser = None

# Flags for lazy imports
__PYTHON_STIX_LOADED = False
__PYTHON_CYBOX_LOADED = False
__PYTHON_MAEC_LOADED = False
__MIXBOX_LOADED = False


def _load_stix():
    global __PYTHON_STIX_LOADED
    global stix, stix_core, stix_campaign, stix_coa, stix_dm, stix_et
    global stix_incident, stix_indicator, stix_report, stix_ta, stix_ttp
    global stix_utils, stix_parser, stix_nsparser, stix_common

    if __PYTHON_STIX_LOADED:
        return

    try:
        import stix
        import stix.core as stix_core
        import stix.campaign as stix_campaign
        import stix.common as stix_common
        import stix.coa as stix_coa
        import stix.data_marking as stix_dm
        import stix.exploit_target as stix_et
        import stix.incident as stix_incident
        import stix.indicator as stix_indicator
        import stix.report as stix_report
        import stix.threat_actor as stix_ta
        import stix.ttp as stix_ttp
        import stix.utils as stix_utils
        import stix.utils.parser as stix_parser
        import stix.utils.nsparser as stix_nsparser
    except (ImportError, ImportWarning) as e:
        LOG.warning(e)
        return

    __PYTHON_STIX_LOADED = True


def _load_cybox():
    global __PYTHON_CYBOX_LOADED
    global cybox, cybox_core, cybox_common, cybox_nsparser

    if __PYTHON_CYBOX_LOADED:
        return

    try:
        import cybox
        import cybox.core as cybox_core
        import cybox.common as cybox_common
        import cybox.utils.nsparser as cybox_nsparser
    except (ImportError, ImportWarning) as e:
        LOG.warning(e)
        return

    __PYTHON_CYBOX_LOADED = True


def _load_mixbox():
    global __MIXBOX_LOADED
    global mixbox, mixbox_entities, mixbox_ns

    if __MIXBOX_LOADED:
        return

    try:
        import mixbox
        import mixbox.entities as mixbox_entities
        import mixbox.namespaces as mixbox_ns
    except (ImportError, ImportWarning) as e:
        LOG.warning(e)
        return

    __MIXBOX_LOADED = True


def _load_maec():
    global __PYTHON_MAEC_LOADED
    global maec, maec_nsparser

    if __PYTHON_MAEC_LOADED:
        return

    try:
        import maec
        import maec.utils.nsparser as maec_nsparser
    except (ImportError, ImportWarning) as e:
        LOG.warning(e)
        return

    __PYTHON_MAEC_LOADED = True


@contextlib.contextmanager
def ignored(*exceptions):
    """Allows you to ignore exceptions cleanly using context managers. This
    exists in Python 3.

    """
    try:
        yield
    except exceptions:
        pass


def fully_qualified_name(object_):
    return get_module(object_) + "." + class_name(object_)


def get_module(object_):
    return object_.__module__


def module_name(module_):
    return str(module_.__name__)


def class_name(object_):
    return str(object_.__class__.__name__)


def handle_typefield(tf):
    if tf.name[0].islower() and tf.type_ is None:

        if tf.name == "classxx":
            return "@{0}".format(tf.key_name)

        elif tf.name == "valueOf_":
            return "text()"

        elif tf.name == "xsi_type":
            return None

        elif tf.name == "type_":
            return "@type"

        return "@{0}".format(tf.name)

    return tf.name


def contains_handling(entity):
    return hasattr(entity, "handling")


def is_stix_core(entity):
    _load_stix()
    return isinstance(entity, stix.BaseCoreComponent)


def is_stix_report(entity):
    _load_stix()
    if not stix_report:  # For python-stix < 1.2.0.0, there is no Report object
        return False
    return isinstance(entity, stix_report.Report)


def is_cybox_property(entity):
    _load_cybox()
    return isinstance(entity, cybox_common.ObjectProperties)


def is_marking_specification(marking):
    _load_stix()
    return isinstance(marking, stix_dm.MarkingSpecification)


def is_marking_structure(marking):
    _load_stix()
    return isinstance(marking, stix_dm.MarkingStructure)


def check_marking(marking):
    if is_marking_specification(marking):
        return

    msg = ("Attempting to mark object with incorrect type: '{0}'. Must be an "
           "instance of MarkingSpecification")
    msg = msg.format(type(marking))
    raise errors.UnknownMarkingError(message=msg, marking=marking)


def check_marking_struct(struct):
    if is_marking_structure(struct):
        return

    msg = "Expected '{0}' but received '{1}'"
    msg.format(stix_dm.MarkingStructure, type(struct))

    raise errors.UnknownMarkingError(message=msg, marking=struct)


def check_empty_marking(marking):
    if not marking.controlled_structure:
        return

    error = "Marking path must be empty. Found '{0}'"
    error = error.format(marking.controlled_structure)

    raise errors.MarkingPathNotEmpty(
        message=error,
        marking=marking
    )


def parse_package(xml_input, encoding=None):
    _load_stix()
    parser = stix_parser.EntityParser()
    return parser.parse_xml(xml_input, encoding=encoding)


def lookup(package, id_):
    entity = package.find(id_)

    if not entity:
        msg = "Failed to find entity in STIX Package with ID {0}"
        msg = msg.format(id_)
        raise errors.IdLookupError(message=msg, id_=id_)

    return entity


def get_handling(entity):
    _load_stix()

    try:
        header = entity.stix_header
        if not header:
            header = stix_core.STIXHeader()
            entity.stix_header = header
        base = header
    except AttributeError:
        # Entity was not a STIXPackage instance.
        base = entity

    if hasattr(base, "handling") and not base.handling:
        base.handling = stix_dm.Marking()

    # Handles stix.Report Handling case.
    elif hasattr(base, "header"):
        if not base.header:
            base.header = stix_report.header.Header()

        if not base.header.handling:
            base.header.handling = stix_dm.Marking()

        return base.header.handling

    return base.handling


def is_package(item):
    _load_stix()
    return isinstance(item, stix_core.STIXPackage)


def is_reference(item):
    try:
        return bool(item.idref)
    except AttributeError:
        return False


def contains_struct(spec, struct):
    structs = spec.marking_structures

    if not structs:
        return False

    if struct in structs:
        return True

    # Try to see if an equivalent struct is in the marking_spec
    struct_dict = struct.to_dict()
    return any(struct_dict == s.to_dict() for s in structs.to_list())


def contains_spec(specs, spec):
    def strip_cs(d):
        """Remove the 'controlled_structure' entry from the dictionary
        if it is present.

        """
        d.pop('controlled_structure', None)
        return d

    if not specs:
        return False

    if spec in specs:
        return True

    spec_dict = strip_cs(spec.to_dict())
    return any(spec_dict == strip_cs(s.to_dict()) for s in specs)


def fields(entity):

    if hasattr(entity, "typed_fields"):
        return entity.typed_fields
    elif hasattr(entity, "_get_vars"):
        return entity._get_vars()
    else:
        return []


def replace_xml_element(old, new):
    """Replaces `old` node with `new` in the document which `old` exists."""
    if old is new:
        return old

    parent = old.getparent()

    if parent is None:
        raise ValueError("Old value has no parent.")

    idx = parent.index(old)
    parent.insert(idx, new)
    parent.remove(old)


def is_entitylist(entity):
    _load_mixbox()
    return isinstance(entity, mixbox_entities.EntityList)


def is_skippable(owner, varname, varobj):

    if varname == "__datamarkings__" and isinstance(varobj, set):
        return True

    if varname == "__datamarkings__" and isinstance(owner, set):
        return True

    if varname == "_fields" and isinstance(varobj, dict):
        return True

    if varname == "_parent" and is_cybox_property(owner):
        return True

    if varname in ("__input_namespaces__", "__input_schemalocations__",
                   "__binding__"):
        return True

    return False


def is_entity(entity):
    _load_mixbox()
    return isinstance(entity, mixbox_entities.Entity)


def is_list(entity):
    return isinstance(entity, list)


def is_node(field):
    """Return True if `field` represents an XML Axis ``self::node()`` or
    ``descendant-or-self::node()`` selector, otherwise False.
    """
    return (xml.XPATH_AXIS_SELF_NODE in field or
            xml.XPATH_AXIS_DESCENDANT_OR_SELF_NODE in field)


def is_sequence(item):
    """Returns ``True`` if `value` is a sequence type (e.g., ``list``, or
    ``tuple``). String types will return ``False``.

    """
    return hasattr(item, "__iter__") and not isinstance(item, string_types)


def attr_name(path, name):
    """Converts `name` TypeField into the attribute name expected by
    MarkingSerializer when walking the object model.

    It will access the last entity instance through `path`, inspect its
    properties and match it against `name` TypeField.

    Arguments:
        name: A TypeField instance.
        path: A `list` that represents a given step while walking the
        python object model.

    """
    for x in path[-1].typed_fields_with_attrnames():
        val, tf = x

        if tf is name:
            return val


def load_nsmap():
    """It combines namespaces to prefix dictionaries of STIX, CybOX, MAEC and
    any other extension that registered their namespace with mixbox.

    Returns:
        Default Namespace to prefix aliases mapping.
    """
    _load_mixbox()

    return mixbox_ns.__ALL_NAMESPACES


def get_all_marking_specs(package):
    """
    This method will walk a STIX Package and return all the MarkingSpecification
    objects found in the input document. The method will only return markings
    that are explicitly set in the input document.

    Args:
        package: stix.core.STIXPackage object.

    Returns:
        list: Containing MarkingSpecification objects.
    """
    from . import navigator

    _load_stix()
    marking_specs = []

    for entity in navigator.iterwalk(package):
        if isinstance(entity, stix_dm.MarkingSpecification):
            marking_specs.append(entity)

    return marking_specs


def get_null_markings(package):
    """
    This method will walk a STIX Package and return all the MarkingSpecification
    objects found in the input document that do NOT contain a Controlled
    Structure. The method will only return markings that are explicitly set in
    the input document.

    Args:
        package: stix.core.STIXPackage object.

    Returns:
        list: Containing MarkingSpecification objects.
    """
    from . import navigator

    _load_stix()
    null_markings = []

    for entity in navigator.iterwalk(package):
        if isinstance(entity, stix_dm.MarkingSpecification):
            if not entity.controlled_structure:
                null_markings.append(entity)

    return null_markings


# ----------------- CLASS DEFINITION -----------------


class ModelMapper(object):
    """The ModelMapper script is not an essential part of the stixmarx library.
     It serves as a complimentary tool that helps automating the creation of
     mappings from the Python Object Model (python-stix, python-cybox,
     python-maec, etc.) that contain the new mixbox changes to XML constructs
     (Element, @attribute or text()). It can be used in any other library
     that implements TypeFields from the mixbox library.

     The generated content is found under stixmarx/fields/ directory.
    """
    __version__ = "0.7"

    def __init__(self, *args):
        self._MAP_PACKAGES = args
        self._FIELDS = {}
        self._VERSIONS = {}
        self._IMPORTS = set()
        self.initialize_mappings()

    def initialize_mappings(self):
        for module_str in self._MAP_PACKAGES:
            imported_module = self._get_module(module_str)
            self._VERSIONS[imported_module.__name__] = imported_module.__version__

            objects = self._get_class_instances(module_name(imported_module), imported_module)
            self._manage_module_objects(objects)

    def _manage_module_objects(self, objects):
        for obj in objects:
            self._parse(obj)

    def _parse(self, object_):
        content = {}
        key = fully_qualified_name(object_)
        import_str = get_module(object_)

        for properties in object_.typed_fields_with_attrnames():
            attr_name, typed_field = properties
            tf = handle_typefield(typed_field)

            if tf:
                content[attr_name] = tf

        if content:
            self._FIELDS[key] = content
            self._IMPORTS.add(import_str)

    @staticmethod
    def _get_module(module_str):
        """Returns a tuple with all objects that could be instantiated
        without arguments.

        Args:
            module_str: A string object. It represents the name of the top-level module.

        Returns:
            A ModuleType object with all sub-modules imported from the given module.
        """
        top_level_module = importlib.import_module(module_str)

        prefix = module_name(top_level_module) + "."

        try:
            for module_loader, name, is_pkg in pkgutil.walk_packages(path=top_level_module.__path__, prefix=prefix):
                importlib.import_module(name, module_str)
        except (ImportError, ImportWarning) as e:
            LOG.debug(str(e))

        return top_level_module

    def _get_class_instances(self, module_str, module):
        """Returns a tuple with objects that could be instantiated
        without arguments.

        Args:
            module: A ModuleType object. It can be either in the lowest or
            highest level module.

        Returns:
            A tuple with `mixbox.Entity` objects from the given module.
        """
        result = ()

        for klass_ in itervalues(vars(module)):
            try:
                if inspect.isclass(klass_) and issubclass(klass_, entities.Entity):
                    result += (klass_(),)
                elif inspect.ismodule(klass_) and (module_str in module_name(klass_) and module_str != module_name(klass_)):
                    result += self._get_class_instances(module_str, klass_)
            except:
                msg = "{0} could not be instantiated without args."
                LOG.debug(msg.format(klass_))
                continue

        return result

    def to_dict(self):
        """
        Returns a dictionary with the default keys.
        Information written as comments in `to_file` will be
        kept in a "metadata" structure.
        """
        return_obj = {
            "metadata": {
                "created": str(datetime.datetime.utcnow()),
                "copyright": "Copyright (c) 2020, The MITRE Corporation. All rights reserved.\n"
                             "See LICENSE.txt for complete terms.\n"
                             "This is an auto-generated file.",
                "description": "Maps Python instance attribute names to XML instance field"
                               " names for Python API classes.",
                "versions": {
                    "stixmarx": version.__version__,
                    "ModelMapper": self.__version__,
                }
            }
        }

        for package, version_str in self._VERSIONS.items():
            return_obj["metadata"]["versions"][package] = version_str

        return_obj["fields"] = copy.deepcopy(self._FIELDS)
        return return_obj

    def to_file(self, location=None):
        """
        Writes content of mapping dictionary to file.

        *Note:
            The output file directory and filename will be
                {Drive}:{PATH}/Desktop/fields.py
        """
        if location is None:
            path = os.path.expanduser("~/Desktop")
            destination = os.path.abspath(path + "/fields.py")
        else:
            destination = location

        with open(destination, "w") as outfile:
            outfile_header(outfile, self._VERSIONS)
            outfile_fields_header(outfile)

            for key in sorted(iterkeys(self._FIELDS)):
                outfile.write(" " * 4)
                outfile.write("\"{0}\": {{\n".format(key))

                for in_key, in_val in sorted(iteritems(self._FIELDS[key])):
                    outfile.write(" " * 8)
                    outfile.write("\"{0}\": \"{1}\",\n".format(in_key, in_val))

                outfile.write(" " * 4)
                outfile.write("},\n")

            outfile.write("}\n")


# ----------------- UTILITIES -----------------


def outfile_header(outfile, versions):
    today = datetime.datetime.utcnow()
    ver_str = str()

    for k, v in iteritems(versions):
        ver_str += "# {mod}\t-\t{ver}\n".format(mod=k, ver=v)

    outfile.write(
        "# Copyright (c) {year}, The MITRE Corporation. All rights reserved.\n"
        "# See LICENSE.txt for complete terms.\n\n"
        "# This is an auto-generated file.\n"
        "{versions}\n"
        "__date__ = \"{datetime}\"\n\n".format(year=today.year,
                                               versions=ver_str,
                                               datetime=today)
    )


def outfile_fields_header(outfile):
    outfile.write(
        "\n# Maps Python instance attribute names to XML instance field names for\n"
        "# Python API classes.\n"
        "_FIELDS = {\n"
    )



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

"""
This module define classes and methods that allow traversal of the object model.
"""

# builtins
import itertools

# external
from mixbox.vendor.six import iteritems

# internal
from stixmarx.utils import is_entitylist, is_skippable, attr_name, is_sequence


class Navigator(object):
    """Enables walking the Python object model. Although, similar to STIX
    iterwalk and iterpath methods, it only walks through mixbox.entities.Entity
    objects.

    The Navigator avoids walking through the __datamarkings__ collection.
    Which may cause an infinite loop in older approaches that walk the
    object model when markings are present.

    Note:
        A Navigator instance should not be created directly. Instead, use
        navigator.iterwalk() or navigator.iterpath() to create the corresponding
        generators.

    """
    def __init__(self):
        pass

    @staticmethod
    def _iter_fields(obj):
        attrs = []

        if hasattr(obj, "_fields"):
            attrs.append(iteritems(obj._fields))

        if hasattr(obj, "__dict__"):
            attrs.append(iteritems(vars(obj)))

        return itertools.chain.from_iterable(attrs)

    def iterwalk(self, entity):
        def yield_and_walk(item):
            if item is None:
                return

            yield item

            for descendant in self.iterwalk(item):
                yield descendant

        for varname, varobj in self._iter_fields(entity):
            if is_skippable(entity, varname, varobj):
                continue

            if is_sequence(varobj) and not is_entitylist(varobj):
                for item in varobj:
                    if is_skippable(varobj, varname, item):
                        continue
                    for descendant in yield_and_walk(item):
                        yield descendant

                continue

            for descendant in yield_and_walk(varobj):
                yield descendant

    def iterpath(self, obj, path=None):
        def yield_and_descend(name, item):
            yield (path, attr_name(path, name), item)

            if item is None:
                return

            for path_info in self.iterpath(item, path):
                yield path_info

        if path is None:
            path = []

        path.append(obj)

        for varname, varobj in self._iter_fields(obj):
            if is_skippable(obj, varname, varobj):
                continue

            if is_sequence(varobj) and not is_entitylist(varobj):
                for item in varobj:
                    for path_info in yield_and_descend(varname, item):
                        yield path_info

            else:
                for path_info in yield_and_descend(varname, varobj):
                    yield path_info

        path.pop()


def iterwalk(entity):
    """Returns an generator which `walks` the input object model. Each
    iteration yields a mixbox.entities.Entity or Python built-in children
    of `entity`.

    This is performed depth-first.

    Args:
        entity: A mixbox.entities.Entity

    Yields:
        Children of `entity`.
    """
    navigator = Navigator()

    for field in navigator.iterwalk(entity):
        yield field


def iterpath(entity):
    """Returns a generator which `walks` the input object model. Each
    iteration yields a triple containing a list of ancestor nodes, the name
    of the field, and the field value.

    This is performed depth-first.

    Args:
        entity: A mixbox.entities.Entity

    Example:
        An Indicator Title with a value of "Test" could be yielded as follows:
        ([STIXPackage, Indicators, Indicator], "title", "Test")

    Yields:
        tuple: Containing three items.
    """
    navigator = Navigator()

    for field in navigator.iterpath(entity):
        yield field



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

"""
This module defines methods for applying and retrieving data markings.
"""

# internal
from stixmarx.api import types
from stixmarx import errors
from stixmarx import utils


_ATTR_DATA_MARKINGS = "__datamarkings__"


def _attach(markable, marking):
    """Attach the MarkingSpecification object `marking` to the `markable` object.

    If `markable` does not have a __datamarkings__ collection, one will be
    created and set.

    Args:
        markable: An object that can accept data marking information.
        marking: A python-stix MarkingSpecification object.
    """
    if not hasattr(markable, _ATTR_DATA_MARKINGS):
        markable.__datamarkings__ = set()
    markable.__datamarkings__.add(marking)


def _assert_markable(m):
    """Assert that the input can have data markings attached to it.

    Args:
        m: An object.

    Raises:
        errors.UnmarkableError: If the input object `m` cannot be marked.
    """
    if is_markable(m):
        return

    error = "Input type %s is not markable." % type(m)
    raise errors.UnmarkableError(error, m)


def is_markable(m):
    """Return True if the input object `m` can be marked.

    Markable objects are:
    * Anything that has a __datamarkings__ attribute.
    * Entity instances (mixbox or python-stix)
    * Builtin scalar types that are castable via the stixmarx.api.types module

    Note:
        Though types.is_castable() will return True for lists that contain
        only castable builtin objects, lists or other iterable collections
        are not markable.

    Args:
        m: An object.

    Returns:
        True if the input object is markable. False otherwise.
    """
    if hasattr(m, _ATTR_DATA_MARKINGS):
        return True
    elif utils.is_entity(m):
        return True
    elif utils.is_sequence(m):
        return False
    else:
        return types.is_castable(m)


def is_marked(markable):
    """Return True if the input `markable` object contains data markings.

    Args:
        markable: An object.
    """
    return bool(getattr(markable, _ATTR_DATA_MARKINGS, False))


def contains_marking(markable, marking):
    """Return True if the `markable` object is marked by the
    MarkingSpecification `marking` object.

    Args:
        markable: A marked object.
        marking: A MarkingSpecification object.
    """
    markings = get_markings(markable)

    if not markings:
        return False

    return marking in markings


def get_markings(markable):
    """Return the collection of data markings attached to the input `markable`
    object.

    Args:
        markable: A marked object.

    Returns:
        None if the object is not marked. A tuple of data markings if `markable`
        is marked.
    """
    if not is_marked(markable):
        return tuple()
    return tuple(markable.__datamarkings__)


def add_markings(markable, markings):
    """Add multiple markings to the `markable` object.

    Note:
        If `markable` is an immutable Python type, the user will need to replace
        the input value with the return value:

        >>> i = Indicator(id_"foo")
        >>> new_id = add_marking(i.id_, MARKING)
        >>> i.id_  = new_id
        >>>
        >>> # OR do it in one line
        >>>
        >>> i.id_ = add_marking(i.id_, MARKING)

    Args:
        markable: A markable object.
        markings: An iterable collection of MarkingSpecification objects.

    Returns:
        The input object (or a markable variant) that has been marked by the
        input `markings`. If `markable` is a Python builtin immutable type, it
        will be coerced into a new stixmarx.api.types type.

    Raises:
        errors.UnmarkableError: If the `markable` object cannot have markings
            attached to it.
    """
    for marking in markings:
        markable = add_marking(markable, marking)

    return markable


def add_marking(markable, marking):
    """Add a single marking to the `markable` object.

    Args:
        markable: A markable object.
        marking: A MarkingSpecification object.

    Returns:
        The input object (or a markable variant) that has been marked by the
        input `marking`. If `markable` is a Python builtin scalar type.

    Raises:
        errors.UnmarkableError: If the `markable` object cannot have markings
            attached to it.
    """
    _assert_markable(markable)

    if types.is_castable(markable):
        markable = types.cast(markable)

    _attach(markable, marking)

    return markable


def remove_markings(markable, markings):
    """Remove the collection of `markings` from the `markable` object. If the
    `markable` object is not marked, do nothing.

    Args:
        markable: A markable object.
        markings: An iterable collection of MarkingSpecification objects.

    Raises:
        KeyError: If `markable` is not marked by one or more of the `markings`.
    """
    for marking in markings:
        remove_marking(markable, marking)


def remove_marking(markable, marking):
    """Remove the `marking` from the `markable` object. If the `markable`
    object is not marked, do nothing.

    Args:
        markable: A markable object.
        marking: A MarkingSpecification object.

    Raises:
        KeyError: If `markable` is not marked by one or more of the `markings`.
    """
    if not is_marked(markable):
        return

    markable.__datamarkings__.remove(marking)


def clear_markings(markable):
    """Remove all marking information from the `markable` object. If `markable`
    is not marked, do nothing.

    Args:
        markable: A markable object.
    """
    if is_marked(markable):
        markable.__datamarkings__ = set()




# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

# stdlib
import logging
import collections

# internal
from stixmarx import xml


# Module-level logger
LOG = logging.getLogger(__name__)

# Required for finding data-marking schema instances.
_NSMAP = {"marking": "http://data-marking.mitre.org/Marking-1"}


class _XmlElement(object):
    """Wraps an etree _Element object.

    This is used to represent XML element objects in a NodeSet collection and
    provides an interface consistent with _XmlAttribute (e.g., they both have
    a `sourcenode` attribute).

    """
    def __init__(self, node):
        self.name = xml.localname(node)
        self.sourcenode = node

    def __eq__(self, other):
        if other is self:
            return True

        try:
            return other.sourcenode is self.sourcenode
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.sourcenode)


class _XmlAttribute(object):
    """Wraps an etree _ElementStringResult or _ElementUnicodeResult
    object which represents an XML attribute.

    An _ElementStringResult object subclasses `str` and a _ElementUnicodeResult
    object subclasses `unicode`, so any two attributes that share the same
    value will be considered equal and have equal hashes even if they belong
    to two different XML elements.

    This class exists to distinguish attributes from each other by factoring
    an attribute's parent into the equality check.

    Note: UTF-8 encoding was selected as default to provide support
        for Unicode. If a consumer utilizes a different character encoding it
        must be provided through stixmarx.parse(xml_input, encoding="...").
    """

    def __init__(self, node, encoding=None):
        self.name = xml.localname(node.attrname)
        self.parent = node.getparent()
        self.sourcenode = node
        self.encoding = "utf-8" if not encoding else encoding

    def __eq__(self, other):
        """Return True if the `other` _XmlAttribute has the same parent
        element, the same attribute name, and the same attribute value.
        """
        if other is self:
            return True

        if other.__class__ != self.__class__:
            return False

        try:
            return (other.sourcenode == self.sourcenode and
                    other.name == self.name and
                    other.parent is self.parent)
        except AttributeError as ex:
            LOG.warn(ex)
            return False

    def __hash__(self):
        return hash((self.name, self.sourcenode, self.parent))


class MarkingMap(collections.MutableMapping):
    def __init__(self, encoding=None):
        self._inner = {}
        self._encoding = encoding

    def __delitem__(self, key):
        return self._inner.__delitem__(self._make_key(key))

    def __len__(self):
        return self._inner.__len__()

    def __iter__(self):
        return self._inner.__iter__()

    def __getitem__(self, key):
        return self._inner.__getitem__(self._make_key(key))

    def _make_key(self, node):
        if isinstance(node, (_XmlAttribute, _XmlElement)):
            return node
        elif xml.is_element(node):
            return _XmlElement(node)
        elif xml.is_attribute(node):
            return _XmlAttribute(node, self._encoding)
        elif xml.is_content(node):
            return _XmlAttribute(node, self._encoding)
        else:
            raise TypeError("Unexpected type: %s", type(node))

    def __contains__(self, item):
        """Override the super() __contains__ which may raise a TypeError from
        _make_key().
        """
        try:
            return super(MarkingMap, self).__contains__(item)
        except TypeError:
            return False

    def __setitem__(self, key, value):
        key = self._make_key(key)
        self._inner[key] = value

    def add(self, key, value):
        key = self._make_key(key)  # Reduce the amount of casting

        if key not in self:
            self[key] = set()

        self[key].add(value)

    def extend(self, key, values):
        key = self._make_key(key)   # Reduce the amount of casting

        if key not in self:
            self[key] = set()

        self[key].update(values)

    def addall(self, keys, value):
        """Add `value` to each key found in `keys`."""
        for key in keys:
            self.add(key, value)


def _get_marking_specifications(root):
    """Find all MarkingSpecificationType instances found inside
    the input document.

    Returns:
        An iterable collection of MarkingSpecificationType lxml Element
        objects.
    """
    xpath = "//marking:Marking"
    return root.xpath(xpath, namespaces=_NSMAP)


def _get_controlled_structure(spec):
    """Return the ``Controlled_Structure`` element found under the input
    `spec` MarkingSpecificationType element.

    Args:
        spec: A MarkingSpecificationType lxml element.

    Returns:
        A single ``Controlled_Structure`` lxml Element node if found. None
        if no ``Controlled_Structure`` is found.
    """
    xpath = "./marking:Controlled_Structure"
    structs = spec.xpath(xpath, namespaces=_NSMAP)

    if len(structs) == 0:
        return None

    return structs[0]


def _get_marked_nodeset(control):
    """Return the nodeset selected by the XPath defined by the input
    Controlled_Structure.

    Args:
        control: A Controlled_Structure lxml Element object.

    Returns:
        An iterable collection of lxml objects.
    """
    xpath = control.text
    namespaces = control.nsmap

    # Fixes empty namespace to prefix key->value pair when default namespace is
    # used. Note that XPath does not have a notion of a default namespace.
    # The empty prefix is therefore undefined for XPath and cannot be used in
    # namespace prefix mappings. The xpath() call can still fail if an invalid
    # xpath expression is provided.
    if None in namespaces:
        del namespaces[None]

    return control.xpath(xpath, namespaces=namespaces)


def build(root, encoding=None):
    """Build a MarkingMap which maps each element and attribute node
    found in the input document to a set of MarkingSpecificationType
    XML instances which marks them.

    The element/attribute nodes are the keys and the set of
    MarkingSpecification Element objects are the values.

    Returns:
        A stixmarx MarkingMap object.
    """
    marked = MarkingMap(encoding)
    specs = _get_marking_specifications(root)

    for spec in specs:
        control = _get_controlled_structure(spec)

        if control is None:
            continue

        nodeset = _get_marked_nodeset(control)
        marked.addall(nodeset, spec)

    return marked



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

from os.path import abspath, dirname, join
from setuptools import setup, find_packages

CUR_DIR = dirname(abspath(__file__))
INIT_FILE = join(CUR_DIR, "stixmarx", "__init__.py")
VERSION_FILE = join(CUR_DIR, "stixmarx", "version.py")


def get_version():
    with open(VERSION_FILE) as f:
        for line in f.readlines():
            if line.startswith("__version__"):
                version = line.split()[-1].strip("\"")
                return version
        raise AttributeError("Package does not have a __version__")


def get_long_description():
    with open("README.rst") as f:
        return f.read()


setup(
    name="stixmarx",
    version=get_version(),
    author="The MITRE Corporation",
    author_email="stix@mitre.org",
    description="A data marking API for STIX 1 content.",
    long_description=get_long_description(),
    long_description_content_type="text/x-rst",
    url="https://github.com/mitre/stixmarx",
    packages=find_packages(),
    install_requires=[
        "stix>=1.1.1.8,<1.2.1.0"
    ],
    license="BSD",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ]
)



# Copyright (c) 2017, The MITRE Corporation. All rights reserved.
# See LICENSE.txt for complete terms.

# builtin
import json
import logging
import os

# internal
from stixmarx import utils

# Module-level logger
LOG = logging.getLogger(__name__)

_FIELD_MAPPINGS = {}


def get_field_mappings():
    return _FIELD_MAPPINGS


def update_field_mappings(mappings):
    global _FIELD_MAPPINGS
    _FIELD_MAPPINGS.update(mappings)


def _initialize_fields():
    utils._load_stix()
    utils._load_cybox()
    utils._load_maec()
    utils._load_mixbox()

    user_path = os.path.join(os.path.expanduser("~"), ".stixmarx")
    if os.path.isdir(user_path) is False:
        os.makedirs(user_path)
        LOG.debug("Created directory '%s'", user_path)

    for loaded_module, module_name in ((utils.stix, "stix"), (utils.cybox, "cybox"), (utils.maec, "maec")):
        if loaded_module:
            fname = module_name + loaded_module.__version__.replace(".", "") + ".json"
            file_location = os.path.join(user_path, fname)

            if os.path.isfile(file_location) is False:
                message = "Generating compatible mappings for %s %s"
                LOG.info(message, module_name, loaded_module.__version__)

                obj_mapper = utils.ModelMapper(module_name)
                model_resource = obj_mapper.to_dict()
                with open(file_location, "w") as f:
                    json.dump(model_resource, f)
                LOG.debug("Saved %s", file_location)
            else:
                with open(file_location, "r") as f:
                    model_resource = json.load(f)
                LOG.debug("Loaded %s", file_location)

            _FIELD_MAPPINGS.update(model_resource["fields"])
            LOG.debug("Updated field mappings for %s", module_name)
        else:
            message = "No %s library found in environment."
            LOG.debug(message, module_name)


_initialize_fields()
